/*
*  cam_ZWO.cpp
*  PHD Guiding
*
*  Created by Robin Glover.
*  Copyright (c) 2014 Robin Glover.
*  All rights reserved.
*
*  This source code is distributed under the following "BSD" license
*  Redistribution and use in source and binary forms, with or without
*  modification, are permitted provided that the following conditions are met:
*    Redistributions of source code must retain the above copyright notice,
*     this list of conditions and the following disclaimer.
*    Redistributions in binary form must reproduce the above copyright notice,
*     this list of conditions and the following disclaimer in the
*     documentation and/or other materials provided with the distribution.
*    Neither the name of Craig Stark, Stark Labs nor the names of its
*     contributors may be used to endorse or promote products derived from
*     this software without specific prior written permission.
*
*  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
*  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
*  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
*  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
*  LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
*  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
*  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
*  INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
*  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
*  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
*  POSSIBILITY OF SUCH DAMAGE.
*
*/
#include "phd.h"

#ifdef ZWO_ASI

#include "cam_ZWO.h"
#include "cameras/ASICamera2.h"

Camera_ZWO::Camera_ZWO()
    : m_buffer(0),
    m_capturing(false)
{
    Connected = false;
    m_hasGuideOutput = true;
    HasSubframes = true;
    HasGainControl = true; // workaround: ok to set to false later, but brain dialog will frash if we start false then change to true later when the camera is connected
}

Camera_ZWO::~Camera_ZWO()
{
    delete[] m_buffer;
}

inline static int cam_gain(int minval, int maxval, int pct)
{
    return minval + pct * (maxval - minval) / 100;
}

inline static int gain_pct(int minval, int maxval, int val)
{
    return (val - minval) * 100 / (maxval - minval);
}

bool Camera_ZWO::Connect()
{
    // Find available cameras
    wxArrayString USBNames;
    int iCameras = ASIGetNumOfConnectedCameras();
    if (iCameras == 0)
    {
        wxMessageBox(_T("No ZWO cameras detected."), _("Error"), wxOK | wxICON_ERROR);
        return true;
    }

    for (int i = 0; i < iCameras; i++)
    {
        ASI_CAMERA_INFO info;
        if (ASIGetCameraProperty(&info, i) == ASI_SUCCESS)
            USBNames.Add(info.Name);
    }

    int iSelected = 0;

    if (USBNames.Count() > 1)
    {
        iSelected = wxGetSingleChoiceIndex(_("Select camera"), _("Camera name"), USBNames);
        if (iSelected == -1)
            return true;
    }

    ASI_CAMERA_INFO info;
    if (ASIGetCameraProperty(&info, iSelected) != ASI_SUCCESS)
    {
        wxMessageBox(_("Failed to get camera properties for ZWO ASI Camera."), _("Error"), wxOK | wxICON_ERROR);
        return true;
    }

    if (ASIOpenCamera(iSelected) != ASI_SUCCESS)
    {
        wxMessageBox(_("Failed to open ZWO ASI Camera."), _("Error"), wxOK | wxICON_ERROR);
        return true;
    }

    m_cameraId = iSelected;
    Connected = true;

    FullSize.x = info.MaxWidth;
    FullSize.y = info.MaxHeight;

    delete[] m_buffer;
    m_buffer = new unsigned char[FullSize.x * FullSize.y];

    PixelSize = info.PixelSize;

    int numControls;
    if (ASIGetNumOfControls(m_cameraId, &numControls) != ASI_SUCCESS)
    {
        Disconnect();
        wxMessageBox(_("Failed to get camera properties for ZWO ASI Camera."), _("Error"), wxOK | wxICON_ERROR);
        return true;
    }

    HasGainControl = false;

    for (int i = 0; i < numControls; i++)
    {
        ASI_CONTROL_CAPS caps;
        if (ASIGetControlCaps(m_cameraId, i, &caps) == ASI_SUCCESS)
        {
            switch (caps.ControlType)
            {
            case ASI_GAIN:
                if (caps.IsWritable)
                {
                    HasGainControl = true;
                    m_gainControlId = caps.ControlID;
                    m_minGain = caps.MinValue;
                    m_maxGain = caps.MaxVale;
                }
                break;
            case ASI_EXPOSURE:
                m_exposureControlId = caps.ControlID;
                break;
            case ASI_BANDWIDTHOVERLOAD:
                ASISetControlValue(m_cameraId, caps.ControlID, caps.MinValue, ASI_FALSE);
                break;
            }
        }

    }

    m_frame = wxRect(FullSize);
    Debug.AddLine("ZWO: frame (%d,%d)+(%d,%d)", m_frame.x, m_frame.y, m_frame.width, m_frame.height);

    ASISetStartPos(m_cameraId, m_frame.GetLeft(), m_frame.GetTop());
    ASISetROIFormat(m_cameraId, m_frame.GetWidth(), m_frame.GetHeight(), 1, ASI_IMG_Y8);

    return false;
}

bool Camera_ZWO::Disconnect()
{
    ASIStopVideoCapture(m_cameraId);
    m_capturing = false;
    ASICloseCamera(m_cameraId);

    Connected = false;

    delete[] m_buffer;
    m_buffer = 0;

    return false;
}

inline static int round_down(int v, int m)
{
    return v & ~(m - 1);
}

inline static int round_up(int v, int m)
{
    return round_down(v + m - 1, m);
}

bool Camera_ZWO::Capture(int duration, usImage& img, wxRect subframe, bool recon)
{
    if (img.Init(FullSize))
    {
        pFrame->Alert(_("Memory allocation error during capture"));
        Disconnect();
        return true;
    }

    wxRect frame;
    wxPoint subframePos; // position of subframe within frame

    bool useSubframe = UseSubframes;

    if (subframe.width <= 0 || subframe.height <= 0)
        useSubframe = false;

    if (useSubframe)
    {
        // ensure transfer size is a multiple of 1024
        //  moving the sub-frame or resizing it is somewhat costly (stopCapture / startCapture)

        frame.SetLeft(round_down(subframe.GetLeft(), 32));
        frame.SetRight(round_up(subframe.GetRight() + 1, 32) - 1);
        frame.SetTop(round_down(subframe.GetTop(), 32));
        frame.SetBottom(round_up(subframe.GetBottom() + 1, 32) - 1);

        subframePos = subframe.GetLeftTop() - frame.GetLeftTop();
    }
    else
    {
        frame = wxRect(FullSize);
    }

    long exposureUS = duration * 1000;
    ASI_BOOL tmp;
    long cur_exp;
    if (ASIGetControlValue(m_cameraId, m_exposureControlId, &cur_exp, &tmp) == ASI_SUCCESS &&
        cur_exp != exposureUS)
    {
        Debug.AddLine("ZWO: set CONTROL_EXPOSURE %d", exposureUS);
        ASISetControlValue(m_cameraId, m_exposureControlId, exposureUS, ASI_FALSE);
    }

    long new_gain = cam_gain(m_minGain, m_maxGain, GuideCameraGain);
    long cur_gain;
    if (ASIGetControlValue(m_cameraId, m_gainControlId, &cur_gain, &tmp) == ASI_SUCCESS &&
        new_gain != cur_gain)
    {
        Debug.AddLine("ZWO: set CONTROL_GAIN %d%% %d", GuideCameraGain, new_gain);
        ASISetControlValue(m_cameraId, m_gainControlId, new_gain, ASI_FALSE);
    }

    bool size_change = frame.GetSize() != m_frame.GetSize();
    bool pos_change = frame.GetLeftTop() != m_frame.GetLeftTop();

    if (size_change || pos_change)
    {
        m_frame = frame;
        Debug.AddLine("ZWO: frame (%d,%d)+(%d,%d)", m_frame.x, m_frame.y, m_frame.width, m_frame.height);

        if (m_capturing)
        {
            Debug.AddLine("ZWO: stopcapture");
            ASIStopVideoCapture(m_cameraId);
            m_capturing = false;

            // clear buffered image?
            ASI_ERROR_CODE status;
            do
            {
                status = ASIGetVideoData(m_cameraId, (unsigned char *) img.ImageData, img.NPixels * sizeof(unsigned short), 0);
                if (status == ASI_SUCCESS)
                    Debug.AddLine("ZWO: getimagedata clearbuf ret %d", status);
            } while (status == ASI_SUCCESS);
        }
    }

    if (size_change)
    {
        ASI_ERROR_CODE status = ASISetROIFormat(m_cameraId, frame.GetWidth(), frame.GetHeight(), 1, ASI_IMG_Y8);
        if (status != ASI_SUCCESS)
            Debug.AddLine("ZWO: setImageFormat(%d,%d) => %d", frame.GetWidth(), frame.GetHeight(), status);
    }

    if (pos_change)
    {
        ASI_ERROR_CODE status = ASISetStartPos(m_cameraId, frame.GetLeft(), frame.GetTop());
        if (status != ASI_SUCCESS)
            Debug.AddLine("ZWO: setStartPos(%d,%d) => %d", frame.GetLeft(), frame.GetTop(), status);
    }

    if (!m_capturing)
    {
        Debug.AddLine("ZWO: startcapture");
        ASIStartVideoCapture(m_cameraId);
        m_capturing = true;
    }

    int frameSize = frame.GetWidth() * frame.GetHeight();

    long timeout = duration * 2 + 15000;
    int poll = wxMin(duration, 100);

    wxStopWatch timer;

    ::wxMilliSleep(duration);

    ASI_ERROR_CODE status;

    do
    {
        status = ASIGetVideoData(m_cameraId, m_buffer, frameSize, poll);
    } while (status != ASI_SUCCESS && timer.Time() <= timeout);

    if (status != ASI_SUCCESS)
    {
        Debug.AddLine("ZWO: getimagedata ret %d", status);
        return true;
    }

    if (useSubframe)
    {
        img.Subframe = subframe;

        // Clear out the image
        memset(img.ImageData, 0, img.NPixels * sizeof(img.ImageData[0]));

        for (int y = 0; y < subframe.height; y++)
        {
            const unsigned char *src = m_buffer + (y + subframePos.y) * frame.width + subframePos.x;
            unsigned short *dst = img.ImageData + (y + subframe.y) * FullSize.GetWidth() + subframe.x;
            for (int x = 0; x < subframe.width; x++)
                *dst++ = *src++;
        }
    }
    else
    {
        for (int i = 0; i < img.NPixels; i++)
            img.ImageData[i] = m_buffer[i];
    }

    if (recon) SubtractDark(img);

    return false;
}

inline static ASI_GUIDE_DIRECTION GetASIDirection(int direction)
{
    switch (direction)
    {
    default:
    case NORTH:
        return ASI_GUIDE_NORTH;
    case EAST:
        return ASI_GUIDE_EAST;
    case WEST:
        return ASI_GUIDE_WEST;
    case SOUTH:
        return ASI_GUIDE_SOUTH;
    }
}

bool Camera_ZWO::ST4PulseGuideScope(int direction, int duration)
{
    ASI_GUIDE_DIRECTION d = GetASIDirection(direction);
    ASIPulseGuideOn(m_cameraId, d);
    wxMilliSleep(duration);
    ASIPulseGuideOff(m_cameraId, d);

    return false;
}

void  Camera_ZWO::ClearGuidePort()
{
    ASIPulseGuideOff(m_cameraId, ASI_GUIDE_NORTH);
    ASIPulseGuideOff(m_cameraId, ASI_GUIDE_SOUTH);
    ASIPulseGuideOff(m_cameraId, ASI_GUIDE_EAST);
    ASIPulseGuideOff(m_cameraId, ASI_GUIDE_WEST);
}

#endif // ZWO_ASI
